from typing import Dict, Optional

from pydantic import BaseModel


class AsyncResponse(BaseModel):
    id: str
    polling_url: str

class SyncResponse(BaseModel):
    id: str
    result: dict
class ImageProcessingResponse(BaseModel):
    task_id: str
    status: str
    result: Optional[Dict] = None
    error: Optional[str] = None



from typing import Optional

from pydantic import Field, model_validator

from blackforest.types.inputs.generic import (
    GenericDimensionInput,
    GenericImageInput,
    GenericImagePromptInput,
)


class FluxProInputs(GenericImageInput,
                    GenericImagePromptInput,
                    GenericDimensionInput):
    steps: Optional[int] = Field(
        default=40,
        ge=1,
        le=50,
        description="Number of steps for the image generation process.",
        example=40,
    )
    guidance: Optional[float] = Field(
        default=2.5,
        ge=1.5,
        le=5.0,
        description="Guidance scale for image generation. \
            High guidance scales improve prompt adherence \
                at the cost of reduced realism.",
        example=2.5,
    )
    interval: Optional[float] = Field(
        default=2.0,
        ge=1.0,
        le=4.0,
        description="Interval parameter for guidance control.",
        example=2.0,
    )

    @model_validator(mode="after")
    def validate_prompt_or_image(self):
        if not self.prompt and not self.image_prompt:
            raise ValueError("Either prompt or image_prompt must be provided")
        return self



from typing import Optional

from pydantic import Field, model_validator

from blackforest.types.inputs.generic import (
    GenericDimensionInput,
    GenericImageInput,
    GenericImagePromptInput,
)


class FluxDevInputs(GenericImageInput,
                    GenericImagePromptInput,
                    GenericDimensionInput):

    steps: Optional[int] = Field(
        default=28,
        ge=1,
        le=50,
        description="Number of steps for the image generation process.",
        example=28,
    )

    guidance: Optional[float] = Field(
        default=3.0,
        ge=1.5,
        le=5.0,
        description="Guidance scale for image generation. \
            High guidance scales improve prompt adherence \
                at the cost of reduced realism.",
        example=3.0,
    )
    safety_tolerance: int = Field(
        default=2,
        ge=0,
        le=6,
        description="Tolerance level for input and output moderation. \
            Between 0 and 6, 0 being most strict, 6 being least strict.",
        example=2,
    )


    @model_validator(mode="after")
    def validate_prompt_or_image(self):
        if not self.prompt and not self.image_prompt:
            raise ValueError("Either prompt or image_prompt must be provided")
        return self




import re
from typing import Optional

from pydantic import Field, HttpUrl, model_validator

from blackforest.types.base.output_format import OutputFormat
from blackforest.types.inputs.generic import GenericImageInput


class FluxKontextProInputs(GenericImageInput):
    """Inputs for the Flux Kontext Pro model."""
    
    # Override prompt to make it required
    prompt: str = Field(
        example="ein fantastisches bild",
        description="Text prompt for image generation.",
    )

    # Override output_format default to png
    output_format: Optional[OutputFormat] = Field(
        default=OutputFormat.png,
        description="Output format for the generated image. Can be 'jpeg' or 'png'.",
    )

    # Kontext-specific input image fields
    input_image: Optional[str] = Field(
        default=None,
        description="Base64 encoded image or URL to use with Kontext.",
    )
    input_image_2: Optional[str] = Field(
        default=None,
        description="Base64 encoded image or URL to use with Kontext. *Experimental Multiref*",
    )
    input_image_3: Optional[str] = Field(
        default=None,
        description="Base64 encoded image or URL to use with Kontext. *Experimental Multiref*",
    )
    input_image_4: Optional[str] = Field(
        default=None,
        description="Base64 encoded image or URL to use with Kontext. *Experimental Multiref*",
    )

    # Custom aspect ratio field with specific validation
    aspect_ratio: Optional[str] = Field(
        default=None, description="Aspect ratio of the image between 21:9 and 9:21"
    )

    @model_validator(mode="after")
    def validate_aspect_ratio(self):
        try:
            if self.aspect_ratio is not None:
                # ensure proper format (1:1) and ratio is between 21:9 and 9:21
                if not re.match(r"^\d+:\d+$", self.aspect_ratio):
                    raise ValueError(
                        "Aspect ratio must be in the format of 'width:height'"
                    )
                width, height = map(int, self.aspect_ratio.split(":"))
                ratio = width / height
                min_ratio = 1 / 4
                max_ratio = 4 / 1
                if not (min_ratio <= ratio <= max_ratio):
                    raise ValueError(
                        f"Aspect ratio {self.aspect_ratio} ({ratio:.3f}) must be between 1:4 and 4:1"
                    )
        except Exception as e:
            raise ValueError(f"Invalid aspect ratio: {e}")
        return self



import os
from typing import Optional

from pydantic import BaseModel, Field, model_validator


# Determine default sync behavior based on environment
def _get_default_sync() -> bool:
    """Get default sync behavior based on environment variables"""
    env = os.environ.get("BFL_ENV", "").lower()
    return env == "dev" or env == "development"


class ClientConfig(BaseModel):
    """Base configuration class for BFL client operations."""
    sync: bool = Field(
        default_factory=_get_default_sync,
        description="Whether to wait for the operation to complete before returning \
            (synchronous mode)."
    )
    timeout: Optional[int] = Field(
        default=None,
        description="Timeout in seconds for synchronous operations. \
            Only used when sync=True."
    )
    polling_interval: float = Field(
        default=1.0,
        ge=0.1,
        description="Time in seconds to wait between polling attempts for \
              synchronous operations."
    )
    max_retries: int = Field(
        default=60,
        ge=1,
        description="Maximum number of polling attempts for synchronous operations."
    )

    @model_validator(mode='after')
    def log_sync_mode(self):
        """Log debug information about sync mode when in development"""
        if os.environ.get("BFL_DEBUG", "").lower() in ("1", "true", "yes"):
            env = os.environ.get("BFL_ENV", "production")
            print(f"[BFL_DEBUG] Environment: {env}, \
                  Sync mode: {'enabled' if self.sync else 'disabled'}")
        return self



from typing import Optional

from pydantic import Field, model_validator

from blackforest.types.inputs.generic import GenericImageInput, GenericImageValidation


class FluxProCannyInputs(GenericImageInput, GenericImageValidation):
    control_image: Optional[str] = Field(
        default=None,
        description="Base64 encoded image to use as control input \
            if no preprocessed image is provided",
    )

    preprocessed_image: Optional[str] = Field(
        default=None,
        description="Optional pre-processed image that will bypass \
            the control preprocessing step",
    )

    canny_low_threshold: Optional[int] = Field(
        default=50, ge=0, le=500, description="Low threshold for Canny edge detection"
    )

    canny_high_threshold: Optional[int] = Field(
        default=200, ge=0, le=500, description="High threshold for Canny edge detection"
    )

    steps: Optional[int] = Field(
        default=50,
        ge=15,
        le=50,
        description="Number of steps for the image generation process",
    )

    guidance: Optional[float] = Field(
        default=30,
        ge=1,
        le=100,
        description="Guidance strength for the image generation process",
    )


    @model_validator(mode="after")
    def validate_control_image(self):
        if not self.control_image and not self.preprocessed_image:
            raise ValueError(
                "Either control_image or preprocessed_image must be provided"
            )

        if self.control_image:
            self._validate_image(self.control_image, field_name="control_image")
        if self.preprocessed_image:
            self._validate_image(
                self.preprocessed_image, field_name="preprocessed_image"
            )
        return self



from pydantic import Field, model_validator

from blackforest.types.inputs.generic import (
    GenericAspectRatioInput,
    GenericImageInput,
    GenericImagePromptInput,
)


class FluxUltraInputs(GenericImageInput,
                      GenericImagePromptInput,
                      GenericAspectRatioInput):
    raw: bool = Field(
        default=False,
        description="Generate less processed, more natural-looking images",
        example=False,
    )

    image_prompt_strength: float = Field(
        default=0.1,
        ge=0,
        le=1,
        description="Blend between the prompt and the image prompt",
    )

    @model_validator(mode="after")
    def validate_prompt_or_image(self):
        if not self.prompt and not self.image_prompt:
            raise ValueError("Either prompt or image_prompt must be provided")
        return self



from pydantic import model_validator

from blackforest.types.inputs.generic import (
    GenericDimensionInput,
    GenericImageInput,
    GenericImagePromptInput,
)


class FluxPro11Inputs(GenericImageInput,
                      GenericImagePromptInput,
                      GenericDimensionInput):
    """Inputs for the Flux Pro 1.1 model."""

    @model_validator(mode="after")
    def validate_prompt_or_image(self):
        if not self.prompt and not self.image_prompt:
            raise ValueError("Either prompt or image_prompt must be provided")
        return self



from typing import Optional

from pydantic import Field, model_validator

from blackforest.types.inputs.generic import GenericImageInput, GenericImageValidation


class FluxProExpandInputs(GenericImageInput, GenericImageValidation):
    image: str = Field(
        description="A Base64-encoded string representing \
            the image you wish to expand.",
    )
    top: Optional[int] = Field(
        default=0,
        ge=0,
        le=2048,
        description="Number of pixels to expand at the top of the image",
    )
    bottom: Optional[int] = Field(
        default=0,
        ge=0,
        le=2048,
        description="Number of pixels to expand at the bottom of the image",
    )
    left: Optional[int] = Field(
        default=0,
        ge=0,
        le=2048,
        description="Number of pixels to expand on the left side of the image",
    )
    right: Optional[int] = Field(
        default=0,
        ge=0,
        le=2048,
        description="Number of pixels to expand on the right side of the image",
    )

    steps: Optional[int] = Field(
        default=50,
        ge=15,
        le=50,
        description="Number of steps for the image generation process",
        example=50,
    )

    guidance: Optional[float] = Field(
        default=60,
        ge=1.5,
        le=100,
        description="Guidance strength for the image generation process",
    )


    @model_validator(mode="after")
    def validate_images(self):
        # Only validate the input image exists and is valid
        image_size, _ = self._validate_image(self.image, field_name="image")

        # Ensure at least one side is being expanded
        if not any([self.top, self.bottom, self.left, self.right]):
            raise ValueError(
                "At least one side must be expanded (top, bottom, left, or right)"
            )

        return self



import base64
import os

import pytest

from blackforest import BFLClient
from blackforest.types.general.client_config import ClientConfig

BFL_API_KEY = os.getenv("BFL_API_KEY", "test-key")
os.environ["BFL_ENV"] = "dev"  # Set environment to dev mode for testing

def test_client_initialization():
    client = BFLClient(api_key="test-key")
    assert client.api_key == "test-key"
    assert client.base_url == "https://api.bfl.ai"
    assert client.timeout == 30

def test_client_custom_base_url():
    client = BFLClient(api_key="test-key", base_url="https://api.bfl.ai")
    assert client.base_url == "https://api.bfl.ai"

def test_client_headers():
    client = BFLClient(api_key="test-key")
    headers = client.session.headers
    assert headers["X-Key"] == "test-key"
    assert headers["Content-Type"] == "application/json"
    assert headers["Accept"] == "application/json"

def test_polling_url_mapping():
    """Test that polling URLs are properly stored and retrieved."""
    client = BFLClient(api_key=BFL_API_KEY)
    
    # Test that mapping starts empty
    assert len(client._task_polling_urls) == 0
    
    # Create a test task
    inputs = {
        "prompt": "test prompt",
        "width": 512,
        "height": 512,
        "output_format": "jpeg"
    }
    
    config = ClientConfig(sync=False)  # Use async to avoid polling
    response = client.generate("flux-pro-1.1", inputs, config)
    
    # Verify that polling URL was stored
    assert response.id in client._task_polling_urls
    stored_url, timestamp = client._task_polling_urls[response.id]
    assert stored_url == response.polling_url
    assert timestamp > 0
    
    # Test that the correct polling endpoint is returned
    endpoint = client._get_polling_endpoint(response.id)
    assert endpoint == response.polling_url
    
    # Test manual cleanup
    client.clear_polling_urls()
    assert len(client._task_polling_urls) == 0

def test_generate_flux_pro_1_1_no_config():
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    # Create input as dictionary instead of model instance
    inputs = {
        "prompt": "a beautiful sunset over mountains, digital art style",
        "width": 1024,
        "height": 768,
        "output_format": "jpeg"
    }

    config = ClientConfig()

    # Call generate with dictionary and config
    response = client.generate("flux-pro-1.1", inputs)
    print(f"Response: {response}")

    if config.sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None
        assert response.polling_url is not None

@pytest.mark.parametrize("model", ["flux-pro-1.1", "flux-pro", "flux-dev"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    # Create input as dictionary instead of model instance
    inputs = {
        "prompt": "a beautiful sunset over mountains, digital art style",
        "width": 1024,
        "height": 768,
        "output_format": "jpeg"
    }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None
        assert response.polling_url is not None

@pytest.mark.parametrize("model", ["flux-pro-1.1-ultra"])
@pytest.mark.parametrize("raw", [True, False])
@pytest.mark.parametrize("aspect_ratio", ["16:9", "9:16"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_ultra_model(model, raw, aspect_ratio, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    # Create input as dictionary instead of model instance
    inputs = {
        "prompt": "a beautiful sunset over mountains, digital art style",
        "width": 1024,
        "height": 768,
        "output_format": "jpeg",
        "raw": raw,
        "aspect_ratio": aspect_ratio,
    }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None
        assert response.polling_url is not None

@pytest.mark.parametrize("model", ["flux-pro-1.0-fill"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_pro_fill_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    image_path = os.path.join(current_dir, "test_images", "image.png")
    mask_path = os.path.join(current_dir, "test_images", "mask.png")

    with open(image_path, "rb") as image_file:
        image_base64 = base64.b64encode(image_file.read()).decode("utf-8")

    with open(mask_path, "rb") as image_file:
        mask_base64 = base64.b64encode(image_file.read()).decode("utf-8")

    # Create input as dictionary instead of model instance
    inputs = {
        "image": image_base64,
        "mask": mask_base64,
        "prompt": "A beautiful landscape with mountains and a lake",
        "prompt_upsampling": True,
        "seed": 42,
        "safety_tolerance": 2,
    }
    if model == "flux-pro-1.0-fill":
        inputs.update({"guidance": 30.0, "steps": 30})

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None

@pytest.mark.parametrize("model", ["flux-pro-1.0-expand"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_pro_expand_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    image_path = os.path.join(current_dir, "test_images", "image.png")

    with open(image_path, "rb") as image_file:
        image_base64 = base64.b64encode(image_file.read()).decode("utf-8")

    # Create input as dictionary instead of model instance
    inputs = {
        "image": image_base64,
        "top": 10,
        "bottom": 0,
        "left": 0,
        "right": 0,
        "prompt": "A beautiful landscape with mountains and a lake",
        "prompt_upsampling": True,
        "seed": 42,
        "safety_tolerance": 2,
    }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None

@pytest.mark.parametrize("model", ["flux-pro-1.0-canny"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_pro_canny_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    image_path = os.path.join(current_dir, "test_images", "image.png")

    with open(image_path, "rb") as image_file:
        image_base64 = base64.b64encode(image_file.read()).decode("utf-8")

    # Create input as dictionary instead of model instance
    inputs = {
        "control_image": image_base64,
        "prompt": "A beautiful landscape with mountains and a lake",
        "steps": 30,
        "prompt_upsampling": True,
        "seed": 42,
        "guidance": 30.0,
        "safety_tolerance": 2,
    }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None


@pytest.mark.parametrize("model", ["flux-pro-1.0-depth"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_pro_depth_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    image_path = os.path.join(current_dir, "test_images", "image.png")

    with open(image_path, "rb") as image_file:
        image_base64 = base64.b64encode(image_file.read()).decode("utf-8")

    # Create input as dictionary instead of model instance
    inputs =  {
            "control_image": image_base64,
            "prompt": "A beautiful landscape with mountains and a lake",
            "steps": 30,
            "prompt_upsampling": True,
            "seed": 42,
            "guidance": 15.0,
            "safety_tolerance": 2,
        }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None


@pytest.mark.parametrize("model", ["flux-kontext-pro"])
@pytest.mark.parametrize("sync", [False, True])
def test_generate_flux_kontext_pro_model(model, sync):
    print(f"Using API key: {BFL_API_KEY}")
    client = BFLClient(api_key=BFL_API_KEY)

    inputs = {
        "prompt": "Make the 2 animals in a scene having breakfast together",
        "input_image": "tests/inputs/test_image_1.jpeg",
        "input_image_2": "tests/inputs/test_image_2.jpeg",
        "aspect_ratio": "16:9",
        "output_format": "png",
        "seed": 42,
        "safety_tolerance": 2,
        "prompt_upsampling": True,
    }

    config = ClientConfig(sync=sync)

    # Call generate with dictionary and config
    response = client.generate(model, inputs, config)
    print(f"Response: {response}")

    if sync:
        assert response.id is not None
        assert response.result is not None
    else:
        assert response.id is not None


from typing import Optional

from pydantic import Field

from blackforest.types.inputs.generic import GenericImageInput


class FluxProFillInputs(GenericImageInput):
    image: str = Field(
        description="A Base64-encoded string representing the image \
            you wish to modify. Can contain alpha mask if desired.",
    )

    mask: Optional[str] = Field(
        default=None,
        description=(
            "A Base64-encoded string representing a mask for the areas you want to \
                modify in the image. "
            "The mask should be the same dimensions as the image \
                and in black and white. Black areas (0%) "
            "indicate no modification, while white areas (100%) \
                specify areas for inpainting. "
            "Optional if you provide an alpha mask in the original image. "
            "Validation: The endpoint verifies that \
                the dimensions of the mask match the original image."
        ),
    )

    steps: Optional[int] = Field(
        default=50,
        ge=15,
        le=50,
        description="Number of steps for the image generation process",
        example=50,
    )

    guidance: Optional[float] = Field(
        default=60,
        ge=1.5,
        le=100,
        description="Guidance strength for the image generation process",
    )



from blackforest.types.inputs.flux_dev import FluxDevInputs
from blackforest.types.inputs.flux_kontext_pro import FluxKontextProInputs
from blackforest.types.inputs.flux_pro import FluxProInputs
from blackforest.types.inputs.flux_pro_1_1 import FluxPro11Inputs
from blackforest.types.inputs.flux_pro_canny import FluxProCannyInputs
from blackforest.types.inputs.flux_pro_depth import FluxProDepthInputs
from blackforest.types.inputs.flux_pro_expand import FluxProExpandInputs
from blackforest.types.inputs.flux_pro_fill import FluxProFillInputs
from blackforest.types.inputs.flux_ultra import FluxUltraInputs

MODEL_INPUT_REGISTRY = {
    "flux-dev": FluxDevInputs,
    "flux-pro": FluxProInputs,
    "flux-pro-1.1": FluxPro11Inputs,
    "flux-pro-1.1-ultra": FluxUltraInputs,
    "flux-pro-1.0-fill": FluxProFillInputs,
    "flux-pro-1.0-expand": FluxProExpandInputs,
    "flux-pro-1.0-canny": FluxProCannyInputs,
    "flux-pro-1.0-depth": FluxProDepthInputs,
    "flux-kontext-pro": FluxKontextProInputs,
}



"""
Main client implementation for the BFL API.
"""

import base64
import os
import time
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin

import requests

from blackforest.resources.mapping.model_input_registry import MODEL_INPUT_REGISTRY
from blackforest.types.general.client_config import ClientConfig
from blackforest.types.inputs.generic import ImageInput
from blackforest.types.responses.responses import (
    AsyncResponse,
    ImageProcessingResponse,
    SyncResponse,
)


class BFLError(Exception):
    """Base exception for BFL API errors."""
    pass

class BFLClient:
    """
    Main client class for interacting with the Black Forest Labs API.
    """

    # Model to input type mapping registry
    model_input_registry = MODEL_INPUT_REGISTRY
    api_version = "v1"
    def __init__(
        self,
        api_key: str,
        base_url: str = "https://api.bfl.ai",
        timeout: int = 30,
    ):
        """
        Initialize the BFL client.

        Args:
            api_key: Your BFL API key
            base_url: Base URL for the API (optional)
            timeout: Request timeout in seconds (optional)
        """
        self.api_key = api_key
        self.base_url = base_url.rstrip('/')
        self.timeout = timeout
        self.session = requests.Session()
        self.session.headers.update({
            'X-Key': api_key,
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        })
        # Map to store task_id -> (polling_url, timestamp)
        self._task_polling_urls = {}

    def _cleanup_expired_polling_urls(self) -> None:
        """
        Remove polling URL entries that are older than 30 minutes.
        """
        current_time = time.time()
        expired_keys = []
        
        for task_id, (polling_url, timestamp) in self._task_polling_urls.items():
            # Remove entries older than 30 minutes (1800 seconds)
            if current_time - timestamp > 1800:
                expired_keys.append(task_id)
        
        for key in expired_keys:
            del self._task_polling_urls[key]

    def clear_polling_urls(self) -> None:
        """
        Manually clear all stored polling URLs.
        """
        self._task_polling_urls.clear()

    def _get_polling_endpoint(self, task_id: str) -> str:
        """
        Get the appropriate polling endpoint for a task.
        
        Args:
            task_id: The task ID to get the endpoint for
            
        Returns:
            The endpoint (relative path or full URL) to use for polling
        """
        # Clean up expired entries
        self._cleanup_expired_polling_urls()
        
        if task_id in self._task_polling_urls:
            polling_url, _ = self._task_polling_urls[task_id]  # Extract URL from tuple
            # Return the full polling URL as-is since _request() now handles full URLs
            return polling_url
        else:
            # Fallback to constructed URL
            return f"/v1/get_result?id={task_id}"

    def _request(
        self,
        method: str,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Make a request to the API.

        Args:
            method: HTTP method
            endpoint: API endpoint (can be a relative path or full URL)
            params: URL parameters
            data: Form data
            json: JSON data

        Returns:
            API response as dictionary

        Raises:
            BFLError: If the API request fails
        """
        # If endpoint is already a full URL, use it directly
        if endpoint.startswith(('http://', 'https://')):
            url = endpoint
        else:
            url = urljoin(self.base_url, endpoint)

        try:
            response = self.session.request(
                method=method,
                url=url,
                params=params,
                data=data,
                json=json,
                timeout=self.timeout,
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            if response is not None:
                try:
                    error_data = response.json()
                    error_message = error_data.get('message', str(e))
                except ValueError:
                    error_message = response.text or str(e)
            else:
                error_message = str(e)

            raise BFLError(f"API request failed: {error_message}") from e

    def _encode_image(self, image_path: str) -> str:
        """Encode image file to base64 string."""
        with open(image_path, 'rb') as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def _is_file_path(self, value: str) -> bool:
        """Check if a string is a file path (not base64 or URL)."""
        if value.startswith(('http://', 'https://')):
            return False
        # Check if it looks like base64 (no path separators, mostly alphanumeric with +/=)
        if '/' not in value and '\\' not in value and len(value) > 100:
            try:
                base64.b64decode(value)
                return False  # It's valid base64
            except Exception:
                pass
        # Check if file exists
        return os.path.exists(value)

    def _process_kontext_inputs(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Process flux-kontext-pro inputs to automatically encode image paths."""
        processed_inputs = inputs.copy()
        
        # Image fields that might need encoding
        image_fields = ['input_image', 'input_image_2', 'input_image_3', 'input_image_4']
        
        for field in image_fields:
            if field in processed_inputs and processed_inputs[field]:
                value = processed_inputs[field]
                if isinstance(value, str) and self._is_file_path(value):
                    try:
                        processed_inputs[field] = self._encode_image(value)
                    except Exception as e:
                        raise BFLError(f"Error encoding image file '{value}' for field '{field}': {str(e)}")
        
        return processed_inputs

    def _process_folder(self, folder_path: str) -> List[str]:
        """Process all images in a folder and return list of base64 encoded images."""
        folder = Path(folder_path)
        if not folder.exists() or not folder.is_dir():
            raise BFLError(f"Invalid folder path: {folder_path}")

        image_extensions = {'.jpg', '.jpeg', '.png', '.gif', '.bmp'}
        encoded_images = []

        for file_path in folder.glob('*'):
            if file_path.suffix.lower() in image_extensions:
                try:
                    encoded_images.append(self._encode_image(str(file_path)))
                except Exception as e:
                    raise BFLError(f"Error processing image {file_path}: {str(e)}")

        return encoded_images

    def _process_zip(self, zip_path: str) -> List[str]:
        """Process all images in a zip file and return list of base64 encoded images."""
        if not os.path.exists(zip_path):
            raise BFLError(f"Invalid zip file path: {zip_path}")

        image_extensions = {'.jpg', '.jpeg', '.png', '.gif', '.bmp'}
        encoded_images = []

        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            for file_name in zip_ref.namelist():
                if Path(file_name).suffix.lower() in image_extensions:
                    try:
                        with zip_ref.open(file_name) as image_file:
                            encoded_images.append(base64.b64encode(image_file.read()).decode('utf-8'))
                    except Exception as e:
                        raise BFLError(f"Error processing image {file_name} \
                                        from zip: {str(e)}")

        return encoded_images

    def process_image(
        self,
        input_data: Union[str, ImageInput],
        endpoint: str = "/v1/image",
        **kwargs
    ) -> ImageProcessingResponse:
        """
        Process an image or multiple images using the specified endpoint.

        Args:
            input_data: Either a path to an image file or an ImageInput object
            endpoint: The API endpoint to use for processing
            **kwargs: Additional parameters to pass to the API

        Returns:
            ImageProcessingResponse containing task ID and status

        Raises:
            BFLError: If there's an error processing the images
        """
        if isinstance(input_data, str):
            # If input_data is a string, assume it's a path to an image
            input_data = ImageInput(image_path=input_data)

        if not isinstance(input_data, ImageInput):
            raise BFLError("input_data must be either a string or an ImageInput object")

        # Process the input based on the provided type
        if input_data.image_path:
            image_data = self._encode_image(input_data.image_path)
        elif input_data.folder_path:
            image_data = self._process_folder(input_data.folder_path)
        elif input_data.zip_path:
            image_data = self._process_zip(input_data.zip_path)
        elif input_data.image_data:
            image_data = input_data.image_data
        else:
            raise BFLError("No valid image input provided")

        # Prepare the request payload
        payload = {
            "image": image_data,
            **kwargs
        }

        # Make the API request
        response = self._request("POST", endpoint, json=payload)

        # Store the polling URL for this task if available
        task_id = response.get("id")
        if task_id and "polling_url" in response:
            # Clean up expired entries before adding new one
            self._cleanup_expired_polling_urls()
            self._task_polling_urls[task_id] = (response["polling_url"], time.time())

        return ImageProcessingResponse(
            task_id=task_id,
            status="submitted",
            result=response
        )

    def get_task_status(self, task_id: str) -> ImageProcessingResponse:
        """
        Get the status of a processing task.

        Args:
            task_id: The ID of the task to check

        Returns:
            ImageProcessingResponse containing current status and result if available
        """
        endpoint = self._get_polling_endpoint(task_id)
        response = self._request("GET", endpoint)

        return ImageProcessingResponse(
            task_id=task_id,
            status=response.get("status", "unknown"),
            result=response.get("result"),
            error=response.get("error")
        )

    def get_polling_result(
        self,
        task_id: str,
        config: Optional[ClientConfig] = None,
    ) -> Dict[str, Any]:
        """
        Poll for results until they are ready or until timeout.

        Args:
            task_id: The ID of the task to poll for
            config: Optional configuration for polling behavior

        Returns:
            The final result from the API

        Raises:
            BFLError: If polling times out or the API request fails
        """
        if config is None:
            config = ClientConfig()

        start_time = time.time()
        attempts = 0

        while attempts < config.max_retries:
            print(f"Polling task {task_id} for result. \
                  Attempt {attempts + 1} of {config.max_retries}")
            
            endpoint = self._get_polling_endpoint(task_id)
            response = self._request("GET", endpoint)

            # Check if the task is complete
            if response.get("status") in ["Ready", "completed", "failed"]:
                if response.get("status") == "failed":
                    error_msg = response.get('error', 'Unknown error')
                    raise BFLError(f"Task failed: {error_msg}")
                return response.get("result", {})

            # Check for timeout
            if config.timeout and (time.time() - start_time > config.timeout):
                raise BFLError(f"Polling timed out after {config.timeout} seconds")

            # Sleep before next attempt
            time.sleep(config.polling_interval)
            attempts += 1

        raise BFLError(f"Polling exceeded maximum retries ({config.max_retries})")

    def generate(
        self,
        model: str,
        inputs: Dict[str, Any],
        config: Optional[ClientConfig] = None,
        track_usage: bool = False,
    ) -> Union[AsyncResponse, SyncResponse, Dict[str, Any]]:
        """
        Generate an image using model

        Args:
            model: The model to use for generation, eg "flux-pro-1.1"
            inputs: Dictionary containing generation parameters
            config: Optional configuration for client behavior
            track_usage: Whether to track usage for licensed models (default: False)

        Returns:
            AsyncResponse containing task ID and polling URL
            OR
            SyncResponse containing actual results

        Raises:
            BFLError: If the API request fails

        Examples:
            >>> from blackforest import BFLClient
            >>> from blackforest.types.general.client_config import ClientConfig
            >>> client = BFLClient(api_key="your-api-key")
            >>>
            >>> # Asynchronous request (default)
            >>> response = client.generate("flux-pro-1.1", \
                                            {"prompt": "a beautiful forest"})
            >>> print(f"Task ID: {response.id}")
            >>>
            >>> # Synchronous request with polling
            >>> config = ClientConfig(sync=True, timeout=120)
            >>> result = client.generate("flux-pro-1.1", \
                                        {"prompt": "a beautiful forest"},
                                        config)
            >>> print(f"Image URL: {result.get('image_url')}")
        """
        if config is None:
            config = ClientConfig()

        # Get the appropriate input type from the registry
        input_cls = self.model_input_registry.get(model)
        if not input_cls:
            raise BFLError(f"Model {model} not supported. \
                           Supported models: {list(self.model_input_registry.keys())}")

        # Process inputs for automatic image encoding if using flux-kontext-pro
        processed_inputs = inputs
        if model == "flux-kontext-pro":
            processed_inputs = self._process_kontext_inputs(inputs)

        # Convert the inputs to the appropriate type
        typed_inputs = input_cls(**processed_inputs)

        response = self._request("POST", f"{self.api_version}/{model}",
                                  json=typed_inputs.model_dump(exclude_none=True))

        # Store the polling URL for this task
        task_id = response["id"]
        if "polling_url" in response:
            # Clean up expired entries before adding new one
            self._cleanup_expired_polling_urls()
            self._task_polling_urls[task_id] = (response["polling_url"], time.time())

        # Track usage if requested
        if track_usage:
            self.track_usage_via_api(model, 1)

        # If sync is True, poll for results
        if config.sync:
            try:
                result = self.get_polling_result(task_id, config)
                return SyncResponse(
                    id=task_id,
                    result=result
                )
            except Exception as e:
                raise BFLError(f"Error getting synchronous result: {str(e)}")

        else:
            return AsyncResponse(
                id=task_id,
                polling_url=response["polling_url"]
            )

    def track_usage_via_api(self, name: str, n: int = 1) -> None:
        """
        Track usage of licensed models via the BFL API for commercial licensing compliance.

        Args:
            name: The model name to track usage for
            n: Number of generations to track (default: 1)

        Raises:
            BFLError: If the API request fails or model is not trackable

        For more information on licensing BFL's models for commercial use and usage reporting,
        see the README.md or visit: https://dashboard.bfl.ai/licensing/subscriptions?showInstructions=true
        """
        model_slug_map = {
            "flux-dev": "flux-1-dev",
            "flux-dev-kontext": "flux-1-kontext-dev",
            "flux-dev-fill": "flux-tools",
            "flux-dev-depth": "flux-tools",
            "flux-dev-canny": "flux-tools",
            "flux-dev-canny-lora": "flux-tools",
            "flux-dev-depth-lora": "flux-tools",
            "flux-dev-redux": "flux-tools",
        }

        if name not in model_slug_map:
            raise BFLError(f"Cannot track usage for model '{name}'. Model not trackable or name incorrect. "
                          f"Trackable models: {list(model_slug_map.keys())}")

        model_slug = model_slug_map[name]
        endpoint = f"/v1/licenses/models/{model_slug}/usage"
        payload = {"number_of_generations": n}

        try:
            self._request("POST", endpoint, json=payload)
            print(f"Successfully tracked usage for {name} with {n} generations")
        except BFLError as e:
            raise BFLError(f"Failed to track usage for {name}: {str(e)}")


import base64
import io
from typing import Optional

from PIL import Image
from pydantic import BaseModel, Field, model_validator

from blackforest.types.base.output_format import OutputFormat


class ImageInput(BaseModel):
    image_path: Optional[str] = Field(None, description="Path to a single image file")
    folder_path: Optional[str] = Field(None, description="Path to a folder \
                                       containing images")
    zip_path: Optional[str] = Field(None, description="Path to a zip file \
                                    containing images")
    image_data: Optional[str] = Field(None, description="Base64 encoded image data")


class GenericAspectRatioInput(BaseModel):
    aspect_ratio: str = Field(
        default="16:9", description="Aspect ratio of the image between 21:9 and 9:21"
    )
class GenericDimensionInput(BaseModel):
    width: int = Field(
        default=1024,
        ge=256,
        le=1440,
        multiple_of=32,
        description="Width of the generated image in pixels. \
            Must be a multiple of 32.",
    )
    height: int = Field(
        default=768,
        ge=256,
        le=1440,
        multiple_of=32,
        description="Height of the generated image in pixels. \
            Must be a multiple of 32.",
    )

class GenericImagePromptInput(BaseModel):
    image_prompt: Optional[str] = Field(
        default=None,
        description="Optional base64 encoded image to use with image models.",
    )

    @model_validator(mode="after")
    def validate_images(self):
        if self.image_prompt is not None:
            # Basic base64 validation
            try:
                import base64
                base64.b64decode(self.image_prompt)
            except Exception:
                raise ValueError("image_prompt must be a valid base64 encoded image")
        return self

class GenericImageValidation:
    @classmethod
    def _validate_image(cls, image: str, field_name: str = "image"):
        if not image:
            raise ValueError(f"{field_name.capitalize()} is required")

        try:
            if len(image) > 20 * 1024 * 1024:
                raise ValueError(
                    f"{field_name.capitalize()} file size exceeds 20MB limit"
                )

            b64_data = base64.b64decode(image)
            img = Image.open(io.BytesIO(b64_data))

            if img.width * img.height > 20 * 10**6:
                raise ValueError(
                    f"{field_name.capitalize()} dimensions exceed 20 megapixels"
                )

            # No side smaller than 256
            if img.width < 256 or img.height < 256:
                raise ValueError(
                    f"{field_name.capitalize()} dimensions \
                        must be at least 256x256 pixels"
                )

            return img.size, img.mode

        except base64.binascii.Error:
            raise ValueError(f"Invalid base64 encoding for {field_name}")
        except ValueError as e:
            raise e
        except Exception:
            raise ValueError(f"Error processing {field_name}")
class GenericImageInput(BaseModel):
    """Base class for image generation inputs."""
    prompt: Optional[str] = Field(
        default="",
        example="ein fantastisches bild",
        description="Text prompt for image generation.",
    )
    seed: Optional[int] = Field(
        default=None,
        description="Optional seed for reproducibility.",
        example=42,
    )
    safety_tolerance: int = Field(
        default=2,
        ge=0,
        le=6,
        description="Tolerance level for input and output moderation. \
            Between 0 and 6, 0 being most strict, 6 being least strict.",
        example=2,
    )
    output_format: Optional[OutputFormat] = Field(
        default=OutputFormat.jpeg,
        description="Output format for the generated image. Can be 'jpeg' or 'png'.",
    )
    prompt_upsampling: bool = Field(
        default=False,
        description="Whether to perform upsampling on the prompt. \
            If active, automatically modifies the prompt for more creative generation.",
    )
    # webhook_url: Optional[HttpUrl] = Field(
    #     default=None, description="URL to receive webhook notifications"
    # )
    # webhook_secret: Optional[str] = Field(
    #     default=None, description="Optional secret for webhook signature verification"
    # )


